//
//  main.c
//  demonbras
//
//  Created by Geoffrey Piekacz on 20/02/2015.
//  Copyright (c) 2015 Geoffrey Piekacz. All rights reserved.
//

/****                           ****/
/** Utility to handle robotic arm **/
/****                           ****/

////
// Include files
////

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <strings.h>
#include <fcntl.h>
#include <termios.h>
#ifdef MODE_UDP
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#endif
#include <libusb-1.0/libusb.h>

////
// Constants
////

#define MAX_DEVICES  16
#define MAX_ENDPOINTS  16

#define CMD_STOP  0x0000
#define CMD_M1_LEFT  0x0001
#define CMD_M1_RIGHT  0x0002
#define CMD_M2_LEFT  0x0004
#define CMD_M2_RIGHT  0x0008
#define CMD_M3_LEFT  0x0010
#define CMD_M3_RIGHT  0x0020
#define CMD_M4_LEFT  0x0040
#define CMD_M4_RIGHT  0x0080
#define CMD_M5_LEFT  0x0100
#define CMD_M5_RIGHT  0x0200
#define CMD_LED_ON  0x0400
#define CMD_LED_OFF  0x0800

#define ARM_TYPE_PEARL  0

#define ARM_PEARL_BUFSIZE_CMD 3
#define ARM_PEARL_REQUEST_TYPE 0x40
#define ARM_PEARL_REQUEST_CMD 0x06
#define ARM_PEARL_VALUE  0x0100
#define ARM_PEARL_INDEX  0x0000
#define ARM_PEARL_CMD_DURATION 200000

#define INTERRUPT_DATA_SIZE 8
#define CONTROL_TIMEOUT  1000
#define INTERRUPT_TIMEOUT 1000
#define MAX_LINE  1024
#define MAX_BUFFER  2048

#define UDP_PORT  5555

////
// Structures
////

typedef struct{
    int vendor;
    int product;
    int type;
} DeviceID;

typedef struct{
    unsigned char type;
    unsigned char address;
} Endpoint;

typedef struct{
    DeviceID id;
    libusb_device_handle *handle;
    Endpoint in[MAX_ENDPOINTS];
    Endpoint out[MAX_ENDPOINTS];
    int nb_in;
    int nb_out;
} Arm;

////
// Global variables
////

DeviceID ID_devices[]={
    {0x1267,0x0000,ARM_TYPE_PEARL},
    {-1,-1}
};

Arm devices[MAX_DEVICES];
int nb_devices=0;
unsigned char led=0;

////
// Fonctions
////

#ifdef MODE_UDP
// Create UDP socket

int initialize_server(int port){
    int s;
    struct sockaddr_in address;
    int status;
    
    s=socket(AF_INET,SOCK_DGRAM,0);
    if(s<0){
        fprintf(stderr,"Could not create socket.\n");
        exit(-1);
    }
    
    address.sin_family=AF_INET;
    address.sin_addr.s_addr=INADDR_ANY;
    address.sin_port=htons(port);
    
    status=bind(s,(struct sockaddr *)&address,sizeof address);
    if(status<0){
        fprintf(stderr,"Could not bind socket to port %d.\n",port);
        return -1;
    }
    
    return s;
}
#endif

// Scan for devices

void scan_for_devices(libusb_context *context){
    libusb_device **list;
    size_t count=libusb_get_device_list(context,&list);
    if(count<0) {perror("scan_for_devices.libusb_get_device_list"); exit(-1);}
    int status;
    size_t i=0;
    for(i=0;i<count;i++){
        libusb_device *device=list[i];
        struct libusb_device_descriptor descriptor;
        status=libusb_get_device_descriptor(device,&descriptor);
        if(status!=0) continue;
        unsigned char isfound=0;
        DeviceID *id=ID_devices;
        while(id->vendor>=0){
            if(descriptor.idVendor==id->vendor &&
               descriptor.idProduct==id->product)
            { isfound=1; break; }
            id++;
        }
        if(isfound){
#ifdef VERBOSE
            uint8_t address=libusb_get_device_address(device);
            printf("Device Found @ Address %d\n",address);
#endif
            if(nb_devices<MAX_DEVICES-1){
                Arm *arm=devices+nb_devices;
                status=libusb_open(device,&(arm->handle));
                if(status!=0){ perror("scan_for_devices.libusb_open"); exit(-1); }
                arm->id=*id;
                arm->nb_in=0;
                arm->nb_out=0;
                nb_devices++;
#ifdef VERBOSE
                printf("We have now %d device(s).\n",nb_devices);
#endif
            }
        }
    }
    libusb_free_device_list(list,1);
}

// Configure devices
// (we use the first configuration and first alternate setting)

void configure_devices(void){
    int status;
    int i,j,k;
    for(i=0;i<nb_devices;i++){
        Arm *arm=devices+i;
        libusb_device_handle *handle=arm->handle;
        libusb_device *device=libusb_get_device(handle);
#ifdef VERBOSE
        printf("Configure arm #%d\n",i);
#endif
        
        /* Detach interfaces if needed */
        struct libusb_config_descriptor *active_cfg_descriptor;
        status=libusb_get_active_config_descriptor(device,&active_cfg_descriptor);
        if(status!=0){
            perror("configure_devices.libusb_get_active_config_descriptor");
            exit(-1);
        }
        for(j=0;j<active_cfg_descriptor->bNumInterfaces;j++){
            const struct libusb_interface *altif_descriptor=
            active_cfg_descriptor->interface+j;
            const struct libusb_interface_descriptor *if_descriptor=
            altif_descriptor->altsetting;
            int if_number=if_descriptor->bInterfaceNumber;
            if(libusb_kernel_driver_active(handle,if_number)){
#ifdef VERBOSE
                printf("  detaching interface #%d from kernel\n",if_number);
#endif
                status=libusb_detach_kernel_driver(handle,if_number);
                if(status!=0)
                { perror("configure_devices.libusb_detach_kernel_driver"); exit(-1); }
            }
        }
        libusb_free_config_descriptor(active_cfg_descriptor);
        
        /* Get the configuration value (first configuration, first alternative) */
        struct libusb_config_descriptor *cfg_descriptor;
        status=libusb_get_config_descriptor(device,0,&cfg_descriptor);
        if(status!=0)
        { perror("configure_devices.libusb_get_config_descriptor"); exit(-1); }
        int cfg_value=cfg_descriptor->bConfigurationValue;
#ifdef VERBOSE
        printf("  using configuration value %d\n",cfg_value);
#endif
        
        /* Set the configuration */
        status=libusb_set_configuration(handle,cfg_value);
        if(status!=0)
        { perror("configure_devices.libusb_set_configuration"); exit(-1); }
#ifdef VERBOSE
        printf("  configuration of value %d selected\n",cfg_value);
#endif
        
        /* Claiming interfaces */
        for(j=0;j<cfg_descriptor->bNumInterfaces;j++){
            const struct libusb_interface *altif_descriptor=cfg_descriptor->interface+j;
            const struct libusb_interface_descriptor *if_descriptor=
            altif_descriptor->altsetting;
            int if_number=if_descriptor->bInterfaceNumber;
            status=libusb_claim_interface(handle,if_number);
            if(status!=0)
            { perror("configure_devices.libusb_claim_interface"); exit(-1); }
#ifdef VERBOSE
            printf("  interface #%d claimed\n",if_number);
#endif
            
            /* Find communication endpoints */
            for(k=0;k<if_descriptor->bNumEndpoints;k++){
                const struct libusb_endpoint_descriptor *ep_descriptor=
                if_descriptor->endpoint+k;
                int type=(ep_descriptor->bmAttributes & 0x3);
                int address=ep_descriptor->bEndpointAddress;
#ifdef VERBOSE
                printf("    endpoint found; type=%02x, address=%02x\n",type,address);
#endif
                unsigned char direction=(address & LIBUSB_ENDPOINT_IN);
                if(direction!=0){
                    int nb=arm->nb_in;
                    arm->in[nb].type=type;
                    arm->in[nb].address=address&(~LIBUSB_ENDPOINT_IN);
                    arm->nb_in++;
                }
                else{
                    int nb=arm->nb_out;
                    arm->out[nb].type=type;
                    arm->out[nb].address=address;
                    arm->nb_out++;
                }
            }
        }
        
        libusb_free_config_descriptor(cfg_descriptor);
    }
}

// Close all ressources

void close_devices(void){
    int status;
    int i,j;
    for(i=0;i<nb_devices;i++){
        libusb_device_handle *handle=devices[i].handle;
#ifdef VERBOSE
        printf("Free ressources for device #%d\n",i);
#endif
        
        /* Get the configuration value */
        libusb_device *device=libusb_get_device(handle);
        struct libusb_config_descriptor *cfg_descriptor;
        status=libusb_get_config_descriptor(device,0,&cfg_descriptor);
        if(status!=0)
        { perror("close_launchers.libusb_get_config_descriptor"); exit(-1); }
        int cfg_value=cfg_descriptor->bConfigurationValue;
#ifdef VERBOSE
        printf("  using configuration value %d\n",cfg_value);
#endif
        
        /* Free interfaces */
        for(j=0;j<cfg_descriptor->bNumInterfaces;j++){
            const struct libusb_interface *altif_descriptor=cfg_descriptor->interface+j;
            const struct libusb_interface_descriptor *if_descriptor=
            altif_descriptor->altsetting;
            int if_number=if_descriptor->bInterfaceNumber;
            status=libusb_release_interface(handle,if_number);
            if(status!=0)
            { perror("close_devices.libusb_release_interface"); exit(-1); }
#ifdef VERBOSE
            printf("  interface #%d unclaimed\n",if_number);
#endif
        }
        
        /* Close device handle */
        libusb_close(handle);
    }
}

// Read interrupts from devices

#ifdef READ_INTERRUPTS
void get_interrupts_from_devices(void){
    int status;
    int i,j,k;
    for(i=0;i<nb_devices;i++){
        Arm *arm=devices+i;
        libusb_device_handle *handle=arm->handle;
        printf("Interrupt(s) from device #%d\n",i);
        for(j=0;j<arm->nb_in;j++){
            int type=arm->in[j].type;
            if(type!=LIBUSB_TRANSFER_TYPE_INTERRUPT) continue;
            int endp=(arm->in[j].address | LIBUSB_ENDPOINT_IN);
            unsigned char data[INTERRUPT_DATA_SIZE];
            int bytes_in;
            printf("  polling interrupt from ep #%02x\n",endp);
            status=libusb_interrupt_transfer(
                                             handle,endp,data,INTERRUPT_DATA_SIZE,&bytes_in,INTERRUPT_TIMEOUT);
            if(status==0 && bytes_in>0){
                printf("  data from ep #%02x => ",endp);
                for(k=0;k<bytes_in;k++) printf("%02x ",data[k]);
                printf("\n");
            }
        }
    }
}
#endif

// Function for controling PEARL robotic arm

void build_message_for_PEARL_arm(int cmd,unsigned char **data,int *size){
    static unsigned char int_data[ARM_PEARL_BUFSIZE_CMD];
    *data=int_data; *size=ARM_PEARL_BUFSIZE_CMD;
    bzero(int_data,ARM_PEARL_BUFSIZE_CMD);
    if((cmd & CMD_M1_LEFT)!=0) int_data[0]=(int_data[0]&~0x02)|0x01;
    if((cmd & CMD_M1_RIGHT)!=0) int_data[0]=(int_data[0]&~0x01)|0x02;
    if((cmd & CMD_M2_LEFT)!=0) int_data[0]=(int_data[0]&~0x08)|0x04;
    if((cmd & CMD_M2_RIGHT)!=0) int_data[0]=(int_data[0]&~0x04)|0x08;
    if((cmd & CMD_M3_LEFT)!=0) int_data[0]=(int_data[0]&~0x20)|0x10;
    if((cmd & CMD_M3_RIGHT)!=0) int_data[0]=(int_data[0]&~0x10)|0x20;
    if((cmd & CMD_M4_LEFT)!=0) int_data[0]=(int_data[0]&~0x80)|0x40;
    if((cmd & CMD_M4_RIGHT)!=0) int_data[0]=(int_data[0]&~0x40)|0x80;
    if((cmd & CMD_M5_LEFT)!=0) int_data[1]=(int_data[1]&~0x02)|0x01;
    if((cmd & CMD_M5_RIGHT)!=0) int_data[1]=(int_data[1]&~0x01)|0x02;
    if(led) int_data[2]=0xff;
#ifdef VERBOSE
    int i;
    printf("Control data: ");
    for(i=0;i<ARM_PEARL_BUFSIZE_CMD;i++) printf("%02hhx ",int_data[i]);
    printf("\n");
#endif
}

void send_command_to_PEARL_arm(Arm *arm,int cmd){
    int type=ARM_PEARL_REQUEST_TYPE;
    int request=ARM_PEARL_REQUEST_CMD;
    int value=ARM_PEARL_VALUE;
    int index=ARM_PEARL_INDEX;
    unsigned char *data;
    int size;
    int nb_bytes;
    build_message_for_PEARL_arm(cmd,&data,&size);
    nb_bytes=libusb_control_transfer(
                                     arm->handle,type,request,value,index,data,size,CONTROL_TIMEOUT);
    if(nb_bytes!=size){
        perror("send_command_to_PEARL_arm.libusb_control_transfer");
        exit(-1);
    }
#ifdef VERBOSE_USB_TRANSFERT
    int k;
    printf("  using control type=%x request=%x  data =>\n    ",type,request);
    for(k=0;k<size;k++) printf("%02x ",data[k]);
    printf("\n");
#endif
    if(cmd!=CMD_STOP && cmd!=CMD_LED_ON && cmd!=CMD_LED_OFF){
        usleep(ARM_PEARL_CMD_DURATION);
        send_command_to_PEARL_arm(arm,CMD_STOP);
    }
}

// Send command to devices

void send_command_to_device(Arm *arm,int cmd){
    switch(arm->id.type){
        case ARM_TYPE_PEARL:
            send_command_to_PEARL_arm(arm,cmd);
            break;
    }
}

void send_command_to_devices(int cmd){
    int i;
    for(i=0;i<nb_devices;i++){
        Arm *arm=devices+i;
#ifdef VERBOSE
        printf("Sending command %02x to arm #%d\n",cmd,i);
#endif
        send_command_to_device(arm,cmd);
    }
}

// Manual control

#ifdef MODE_MANUAL
void read_user_commands(void){
    char line[MAX_LINE];
    while(fgets(line,MAX_LINE-1,stdin)){
        char cmd[MAX_LINE];
        sscanf(line,"%[a-z0-9 +-]",cmd);
        if(strcmp(cmd,"quit")==0) break;
        int code=CMD_STOP;
        if(strstr(cmd,"m1+")!=NULL) code |= CMD_M1_RIGHT;
        if(strstr(cmd,"m1-")!=NULL) code |= CMD_M1_LEFT;
        if(strstr(cmd,"m2+")!=NULL) code |= CMD_M2_RIGHT;
        if(strstr(cmd,"m2-")!=NULL) code |= CMD_M2_LEFT;
        if(strstr(cmd,"m3+")!=NULL) code |= CMD_M3_RIGHT;
        if(strstr(cmd,"m3-")!=NULL) code |= CMD_M3_LEFT;
        if(strstr(cmd,"m4+")!=NULL) code |= CMD_M4_RIGHT;
        if(strstr(cmd,"m4-")!=NULL) code |= CMD_M4_LEFT;
        if(strstr(cmd,"m5+")!=NULL) code |= CMD_M5_RIGHT;
        if(strstr(cmd,"m5-")!=NULL) code |= CMD_M5_LEFT;
        if(strstr(cmd,"on")!=NULL) led=1;
        if(strstr(cmd,"off")!=NULL) led=0;
        if(code==0) printf("Bad command\n");
        else{
#ifdef DEBUG
            printf("Sending command %02x to devices\n",code);
#endif
            send_command_to_devices(code);
#ifdef READ_INTERRUPTS
            get_interrupts_from_devices();
#endif
        }
    }
}
#endif

#ifdef MODE_UDP
void read_udp_commands(int socket){
    while(1){
        struct sockaddr_in address;
        socklen_t size=sizeof address;
        char usercmd[MAX_BUFFER];
        int code=CMD_STOP;
        int userlen=recvfrom(socket,usercmd,MAX_BUFFER-1,0,
                             (struct sockaddr *)&address,&size);
        if(userlen<=0) break;
        switch(usercmd[0]){
            case '0': led=0; break;
            case '1': led=1; break;
            case 'a': code=CMD_M1_LEFT; break;
            case 'A': code=CMD_M1_RIGHT; break;
            case 'b': code=CMD_M2_LEFT; break;
            case 'B': code=CMD_M2_RIGHT; break;
            case 'c': code=CMD_M3_LEFT; break;
            case 'C': code=CMD_M3_RIGHT; break;
            case 'd': code=CMD_M4_LEFT; break;
            case 'D': code=CMD_M4_RIGHT; break;
            case 'e': code=CMD_M5_LEFT; break;
            case 'E': code=CMD_M5_RIGHT; break;
        }
#ifdef DEBUG
        printf("Sending command %02x to devices\n",code);
#endif
        send_command_to_devices(code);
    }
}
#endif

// Main procedure

int main(void){
    libusb_context *context;
    int status=libusb_init(&context);
    if(status!=0){ perror("libusb_init"); exit(-1); }
    scan_for_devices(context); 
    configure_devices();
#ifdef MODE_MANUAL
    read_user_commands();
#endif
#ifdef MODE_UDP
    int socket;
    if((socket=initialize_server(UDP_PORT))<0){
        fprintf(stderr,"Can not initialize UDP server, aborting.\n");
        exit(-1);
    }
    read_udp_commands(socket);
#endif
    close_devices();
    libusb_exit(context);
    exit(0);
}
